/*!
  * @license MIT (https://github.com/geotrev/omdomdom/blob/master/LICENSE)
  * omdomdom v0.1.13 (https://github.com/geotrev/omdomdom)
  * Copyright 2021 George Treviranus <geotrev@icloud.com>
  */
'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var DATA_KEY_ATTRIBUTE$1 = "data-key";
/**
 * Checks if the object property exists.
 * @param {Object.<string, any>} obj
 * @param {string} prop
 * @returns {boolean}
 */

var hasProperty = function hasProperty(obj, prop) {
  return Object.prototype.hasOwnProperty.call(obj, prop);
};
/**
 * Checks if a property in a keyMap is unique. if not, it's skipped.
 * @param {Object.<string, VirtualNode>} map
 * @param {string} key
 * @returns {boolean}
 */

var keyIsValid = function keyIsValid(map, key) {
  if (!key) return false;

  if (hasProperty(map, key)) {
    // eslint-disable-next-line no-console
    console.warn("[OmDomDom]: Children with duplicate keys detected. Children with duplicate keys will be skipped, resulting in dropped node references. Keys must be unique and non-indexed.");
    return false;
  }

  return true;
};
/**
 * lodash forEach, basically.
 * If `false` is explicitly returned, break the loop.
 * @param {[]} items
 * @param fn
 */


var forEach = function forEach(items, fn) {
  var length = items.length;
  var idx = -1;
  if (!length) return;

  while (++idx < length) {
    if (fn(items[idx], idx) === false) break;
  }
};
/**
 * Generates a <key: vNode> map of a virtual node list.
 * @param {VirtualNode[]} children
 * @returns {Object.<string, VirtualNode>}
 */

var createKeyMap = function createKeyMap(children) {
  var map = {};
  var hasKeys = false;
  forEach(children, function (child) {
    var key = child.attributes[DATA_KEY_ATTRIBUTE$1];

    if (keyIsValid(map, key)) {
      map[key] = child;
      if (!hasKeys) hasKeys = true;
    }
  });
  return hasKeys ? map : null;
};
var insertBefore = function insertBefore(parent, child, refNode) {
  return parent.node.insertBefore(child.node, refNode);
};
/**
 * Transfers properties from one virtual node to another.
 * @param {VirtualNode} template
 * @param {VirtualNode} vNode
 */

var patch = function patch(template, vNode) {
  for (var property in template) {
    vNode[property] = template[property];
  }
};

/**
 * Convert stringified HTML into valid HTML, stripping all extra spaces.
 * @param {string} stringToRender
 */
var toHTML = function toHTML(stringToRender) {
  /**
   * Remove all extraneous whitespace:
   * - From the beginning + end of the document fragment
   * - If there's more than one space before a left tag bracket, replace them with one
   * - If there's more than one space before a right tag bracket, replace them with one
   */
  var processedDOMString = stringToRender.trim().replace(/\s+</g, "<").replace(/>\s+/g, ">");
  var parser = new DOMParser();
  var context = parser.parseFromString(processedDOMString, "text/html");
  return context.body;
};

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _unsupportedIterableToArray(arr, i) || _nonIterableRest();
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArrayLimit(arr, i) {
  if (typeof Symbol === "undefined" || !(Symbol.iterator in Object(arr))) return;
  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

var Types = {
  STRING: "string",
  INT: "number",
  BOOL: "boolean"
};
var DomProperties = {};
/**
 * Creates a new record for DomProperties
 * @param {string} attrName
 * @param {string} propName
 * @param {string} type
 */

var createRecord = function createRecord(attrName, propName, type) {
  return {
    attrName: attrName,
    propName: propName,
    type: type
  };
}; // Set string records


var stringProps = [// Style is a special snowflake, and needs an extra property to set its text
// value. This is less confusing and achieves the same result with one reflow,
// compared to a new reflow for each Element.style[styleProperty] setter.
["style", "cssText"], ["class", "className"]];
forEach(stringProps, function (_ref) {
  var _ref2 = _slicedToArray(_ref, 2),
      attr = _ref2[0],
      property = _ref2[1];

  DomProperties[attr] = createRecord(attr, property || attr, Types.STRING);
}); // Set boolean records

var booleanProps = ["autofocus", "draggable", "hidden", "checked", "multiple", "muted", "selected"];
forEach(booleanProps, function (attr) {
  DomProperties[attr] = createRecord(attr, attr, Types.BOOL);
}); // Set Integer records

var integerProps = [["tabindex", "tabIndex"]];
forEach(integerProps, function (_ref3) {
  var _ref4 = _slicedToArray(_ref3, 2),
      attr = _ref4[0],
      property = _ref4[1];

  DomProperties[attr] = createRecord(attr, property, Types.INT);
});
/**
 * Attributes with these namespaces require `setAttributeNS`
 * @type {Object.<string, string>}
 */

var Namespace = {
  xlink: {
    prefix: "xlink:",
    resource: "http://www.w3.org/1999/xlink"
  },
  xml: {
    prefix: "xml:",
    resource: "http://www.w3.org/XML/1998/namespace"
  }
};

/**
 * Set a given attribute as a property, if it has a property equivalent
 * @param {HTMLElement} node
 * @param {string|number|boolean} type
 * @param {string} prop
 * @param {*} value
 */

var setProperty = function setProperty(element, type, prop, value) {
  switch (type) {
    case Types.STRING:
      {
        if (prop === DomProperties.style.propName) {
          if (value === null) {
            element.style[prop] = "";
          } else {
            element.style[prop] = value;
          }
        } else if (value === null) {
          element[prop] = "";
        } else {
          element[prop] = value;
        }

        break;
      }

    case Types.INT:
      {
        if (value === null) {
          var attr = prop.toLowerCase(); // The attribute needs to be removed to reset the property,
          // otherwise the value coerced + reflected onto the attribute.

          element.removeAttribute(attr);
        } else if (value === "0") {
          element[prop] = 0;
        } else if (value === "-1") {
          element[prop] = -1;
        } else {
          var parsed = parseInt(value, 10);

          if (!isNaN(parsed)) {
            element[prop] = parsed;
          }
        }

        break;
      }

    case Types.BOOL:
      {
        if (["", "true"].indexOf(value) < 0) {
          element[prop] = false;
        } else {
          element[prop] = true;
        }

        break;
      }
  }
};

/**
 * Removes stale attributes from the element.
 * @param {HTMLElement} vNode
 * @param {string[]} attributes
 */

var removeAttributes = function removeAttributes(vNode, attributes) {
  forEach(attributes, function (attrName) {
    if (hasProperty(DomProperties, attrName)) {
      var propertyRecord = DomProperties[attrName];
      setProperty(vNode.node, propertyRecord.type, propertyRecord.propName, null);
    } else {
      if (attrName in vNode.node) {
        setProperty(vNode.node, Types.STRING, attrName, null);
      }

      vNode.node.removeAttribute(attrName);
    }

    delete vNode.attributes[attrName];
  });
};
/**
 * Adds attributes to the element.
 * @param {VirtualNode} vNode
 * @param {Object.<string, string>} attributes
 */


var setAttributes = function setAttributes(vNode, attributes) {
  for (var attrName in attributes) {
    var value = attributes[attrName];
    vNode.attributes[attrName] = value;

    if (hasProperty(DomProperties, attrName)) {
      var propertyRecord = DomProperties[attrName];
      setProperty(vNode.node, propertyRecord.type, propertyRecord.propName, value);
      continue;
    } // Set namespaced properties using setAttributeNS


    if (attrName.startsWith(Namespace.xlink.prefix)) {
      vNode.node.setAttributeNS(Namespace.xlink.resource, attrName, value);
      continue;
    }

    if (attrName.startsWith(Namespace.xml.prefix)) {
      vNode.node.setAttributeNS(Namespace.xml.resource, attrName, value);
      continue;
    }

    if (attrName in vNode.node) {
      setProperty(vNode.node, Types.STRING, attrName, value);
    }

    vNode.node.setAttribute(attrName, value || "");
  }
};
/**
 * Checks property-based attributes. If the attribute exists,
 * then we should set its property value.
 * @param {HTMLElement} element
 * @param {Object.<string, string>} attributes
 */


var getPropertyValues = function getPropertyValues(element, attributes) {
  for (var attrName in DomProperties) {
    var propertyRecord = DomProperties[attrName];
    var propName = propertyRecord.propName;
    var attrValue = element.getAttribute(attrName);

    if (attrName === DomProperties.style.attrName) {
      attributes[attrName] = element.style[propName];
    } else if (typeof attrValue === "string") {
      attributes[attrName] = attrValue;
    }
  }
};
/**
 * Gets non-property attributes.
 * @param {HTMLElement} element
 * @returns {Object.<string, string>}
 */


var getBaseAttributes = function getBaseAttributes(element) {
  return Array.prototype.reduce.call(element.attributes, function (attributes, attrName) {
    if (!hasProperty(DomProperties, attrName.name)) {
      attributes[attrName.name] = attrName.value;
    }

    return attributes;
  }, {});
};
/**
 * Gets all attributes.
 * @param {HTMLElement} element
 * @returns {Object.<string, string>}
 */


var getAttributes = function getAttributes(element) {
  var attributes = getBaseAttributes(element);
  getPropertyValues(element, attributes);
  return attributes;
};
/**
 * Reconcile attributes from vNode to template
 * @param {VirtualNode} template
 * @param {VirtualNode} vNode
 */

var updateAttributes = function updateAttributes(template, vNode) {
  var removedAttributes = [];
  var changedAttributes = {}; // Get stale attributes

  for (var attrName in vNode.attributes) {
    var oldValue = vNode.attributes[attrName];
    var nextValue = template.attributes[attrName];
    if (oldValue === nextValue) continue;

    if (typeof nextValue === "undefined") {
      removedAttributes.push(attrName);
    }
  } // Get changed or new attributes


  for (var _attrName in template.attributes) {
    var _oldValue = vNode.attributes[_attrName];
    var _nextValue = template.attributes[_attrName];
    if (_oldValue === _nextValue) continue;

    if (typeof _nextValue !== "undefined") {
      changedAttributes[_attrName] = _nextValue;
    }
  } // Add and remove attributes


  removeAttributes(vNode, removedAttributes);
  setAttributes(vNode, changedAttributes);
};

var DATA_KEY_ATTRIBUTE = "data-key";
/**
 * Both template and vNode have arrays of virtual nodes. Diff them.
 * @param {VirtualNode} template - new virtual node tree.
 * @param {VirtualNode} vNode - existing virtual node tree.
 */

var updateChildren = function updateChildren(template, vNode, update) {
  var templateChildrenLength = template.children.length;
  var vNodeChildrenLength = vNode.children.length;
  if (!templateChildrenLength && !vNodeChildrenLength) return;
  var vNodeKeyMap = createKeyMap(vNode.children); // There were no keys found:

  if (!vNodeKeyMap) {
    // Remove extra nodes if template.children is smaller
    var _delta = vNodeChildrenLength - templateChildrenLength;

    if (_delta > 0) {
      while (_delta-- > 0) {
        var child = vNode.children.pop();
        vNode.node.removeChild(child.node);
      }
    }

    return forEach(template.children, function (templateChild, idx) {
      var vNodeChild = vNode.children[idx];

      if (typeof vNodeChild !== "undefined") {
        update(templateChild, vNodeChild);
      } else {
        vNode.node.appendChild(templateChild.node);
        vNode.children.push(templateChild);
      }
    });
  } // There were keys found, resolve them:


  var nextChildren = Array(templateChildrenLength); // Match keys and update/move children in-place

  forEach(template.children, function (child, idx) {
    var childNodes = vNode.node.childNodes;
    var key = child.attributes[DATA_KEY_ATTRIBUTE];

    if (Object.prototype.hasOwnProperty.call(vNodeKeyMap, key)) {
      var keyChild = vNodeKeyMap[key];

      if (Array.prototype.indexOf.call(childNodes, keyChild.node) !== idx) {
        insertBefore(vNode, keyChild, childNodes[idx]);
      }

      nextChildren[idx] = keyChild; // Prevent duplicates, remove the entry and let it insert at
      // its natural index in the `else` block.

      delete vNodeKeyMap[key];
      update(child, nextChildren[idx]);
    } else {
      insertBefore(vNode, child, childNodes[idx]);
      nextChildren[idx] = child;
    }
  });
  vNode.children = nextChildren; // Remove any real nodes that are left over from the diff

  var childNodesLength = vNode.node.childNodes.length;
  var delta = childNodesLength - templateChildrenLength;

  if (delta > 0) {
    while (delta-- > 0) {
      vNode.node.removeChild(vNode.node.childNodes[childNodesLength - 1]);
      childNodesLength--;
    }
  }
};

/**
 * Object representation of a DOM element.
 * @typedef VirtualNode
 * @type {Object}
 * @property {string} type - The type of node. E.g., tag name, "comment", or "text".
 * @property {Object.<string, string>} attributes - List of attributes, if any.
 * @property {TextNode|CommentNode} content - Text or comment content, if any.
 * @property {VirtualNode[]} children - Child virtual nodes, if any.
 * @property {HTMLElement} node - The corresponding DOM element.
 * @property {boolean} isSVGContext
 */

/**
 * Reconcile differences between two virtual DOM trees.
 * @param {VirtualNode} template - new virtual node tree.
 * @param {VirtualNode} vNode - existing virtual node tree.
 * @param {Node} rootNode - the HTML element containing the current node context
 */

var update = function update(template, vNode) {
  var rootNode = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : vNode.node.parentNode;
  // Node nodes to compare, exit
  if (!template && !vNode) return;
  var contentChanged = template.content && template.content !== vNode.content; // If the type or content changed, replace the node completely

  if (template.type !== vNode.type || contentChanged) {
    rootNode.replaceChild(template.node, vNode.node);
    return patch(template, vNode);
  } // Update attributes, if any


  updateAttributes(template, vNode); // Diff child nodes recursively

  updateChildren(template, vNode, update);
};
/**
 * Renders a node into the given root context. This happens one time,
 * when a component is first rendered.
 * All subsequent renders are the result of reconciliation.
 * @param {VirtualNode} vDOM
 * @param {ShadowRoot|HTMLElement} root
 */

var render = function render(vNode, root) {
  root.appendChild(vNode.node);
};
/**
 * Creates a new virtual DOM from a root node.
 * @param {HTMLElement|ShadowRoot|HTMLBodyElement} node
 * @param {boolean} isSVGContext
 * @returns {VirtualNode}
 */

var create = function create(node) {
  var isSVGContext = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;

  if (typeof node === "string") {
    node = toHTML(node);
  }

  var isRoot = node.tagName === "BODY";
  var childNodes = node.childNodes;
  var numChildNodes = childNodes ? childNodes.length : 0; // toHTML returns a `body` tag as its root node, but we want the first child only

  if (isRoot) {
    if (numChildNodes > 1) {
      throw new Error("[OmDomDom]: Your element should not have more than one root node.");
    } else if (numChildNodes === 0) {
      throw new Error("[OmDomDom]: Your element should have at least one root node.");
    } else {
      return create(childNodes[0]);
    }
  }

  var type = node.nodeType === 3 ? "text" : node.nodeType === 8 ? "comment" : node.tagName.toLowerCase();
  var isSVG = isSVGContext || type === "svg";
  var attributes = node.nodeType === 1 ? getAttributes(node) : {};
  var content = numChildNodes > 0 ? null : node.textContent; // Recursively build children

  var children = Array(numChildNodes);
  forEach(childNodes, function (child, idx) {
    children[idx] = create(child, isSVG);
  });
  return {
    type: type,
    attributes: attributes,
    children: children,
    content: content,
    node: node,
    isSVGContext: isSVG
  };
};

exports.create = create;
exports.render = render;
exports.update = update;
//# sourceMappingURL=omdomdom.cjs.js.map
